from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Number")


@dataclass
class Number(Property):
    number: int | None
    TYPE: str = "number"

    def __init__(
        self,
        name: str,
        id: str | None = None,  # noqa: A002
        number: int | None = None,
    ) -> None:
        self.name = name
        self.id = id
        self.number = number

    def add(self, count: int):
        prev = self.number if self.number is not None else 0
        cls = type(self)
        return cls(
            name=self.name,
            id=self.id,
            number=prev + count,
        )

    @classmethod
    def of(cls: Type[T], name: str, param: dict) -> T:
        if param["number"] is None:
            return cls(name=name, id=param["id"])
        return cls(
            name=name,
            id=param["id"],
            number=param["number"],
        )

    @classmethod
    def empty(cls: Type[T], name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME)

    def is_empty(self) -> bool:
        return self.number is None

    @property
    def value(self) -> int:
        return self.number or 0

    def __dict__(self) -> dict:
        result = {
            "type": self.TYPE,
            self.TYPE: self.number,
        }
        if self.id is not None:
            result["id"] = self.id
        return {
            self.name: result,
        }

    @classmethod
    def from_num(cls: Type[T], value: int, name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            number=value,
        )

    @property
    def _prop_type(self) -> Prop:
        return Prop.NUMBER

    @property
    def _value_for_filter(self) -> Any:
        if self.number is None:
            raise ValueError(f"{self.name}: Number is required")
        return self.number



from abc import ABCMeta, abstractmethod

from .prop import Prop


class Property(metaclass=ABCMeta):
    id: str | None
    name: str
    TYPE: str = "Property"  # Must be overridden
    PROP_NAME: str = "Property"  # Must be overridden

    @abstractmethod
    def __dict__(self) -> dict:
        pass

    @property
    @abstractmethod
    def _prop_type(self) -> Prop:
        pass

    @property
    @abstractmethod
    def _value_for_filter(self):  # noqa: ANN201
        pass



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from ..block.rich_text import RichText
from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Text")


@dataclass
class Text(Property):
    rich_text: RichText
    TYPE: str = "rich_text"

    def __init__(
        self,
        name: str,
        rich_text: RichText,
        id: str | None = None,
    ) -> None:  # noqa: A002
        self.name = name
        self.id = id
        self.rich_text = rich_text

    @classmethod
    def from_dict(cls: Type[T], name: str, param: dict) -> T:
        try:
            rich_text = RichText.from_entity(param["rich_text"])
            id = param["id"]
            return cls(
                name=name,
                id=id,
                rich_text=rich_text,
            )
        except Exception as e:
            print(param)
            raise e

    def __dict__(self):
        result = {
            "type": self.TYPE,
            "rich_text": self.rich_text.to_dict(),
        }
        return {self.name: result}

    def append_text(self, text: str):
        updated_text = self.text + "\n" + text
        cls = self.__class__
        return cls(
            name=self.name,
            rich_text=RichText.from_plain_text(updated_text.strip()),
        )

    @classmethod
    def from_plain_text(cls: Type[T], text: str, name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=RichText.from_plain_text(text=text),
        )

    @classmethod
    def from_rich_text(cls: Type[T], rich_text: RichText, name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME, rich_text=rich_text)

    @classmethod
    def empty(cls: Type[T], name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=RichText.empty(),
        )

    @property
    def text(self) -> str:
        return self.rich_text.to_plain_text()

    @property
    def _prop_type(self) -> Prop:
        return Prop.RICH_TEXT

    @property
    def _value_for_filter(self) -> Any:
        return self.text



from .property import Property
from .properties import Properties
from .date import Date
from .title import Title
from .url import Url
from .relation import Relation
from .status import Status
from .text import Text
from .multi_select import MultiSelect, MultiSelectElement, MultiSelectElements
from .select import Select, Selects
from .checkbox import Checkbox
from .number import Number
from .cover import Cover
from .icon import Icon
from .parent import Parent
from .rollup import Rollup
from .person import People
from .email import Email
from .phone_number import PhoneNumber
from .created_by import CreatedBy
from .last_edited_by import LastEditedBy
from .created_time import CreatedTime
from .last_edited_time import LastEditedTime
from .button import Button
from .formula import Formula
from .unique_id import UniqueId
from .files import Files

__all__ = [
    "Property",
    "Properties",
    "Date",
    "Title",
    "Url",
    "Relation",
    "Status",
    "Text",
    "MultiSelect",
    "MultiSelectElement",
    "MultiSelectElements",
    "Select",
    "Selects",
    "Checkbox",
    "Number",
    "Cover",
    "Icon",
    "Parent",
    "Rollup",
    "People",
    "Email",
    "PhoneNumber",
    "CreatedBy",
    "LastEditedBy",
    "CreatedTime",
    "LastEditedTime",
    "Button",
    "MultiSelect",
    "Formula",
    "UniqueId",
    "Files",
]



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .property import Property

T = TypeVar("T", bound="Button")


@dataclass
class Button(Property):
    id: str
    name: str
    TYPE: str = "button"

    @classmethod
    def of(cls: Type[T], key: str, property: dict) -> T:
        return cls(id=property["id"], name=key)

    def value_for_filter(self) -> str:
        raise NotImplementedError

    def __dict__(self) -> dict:
        return {
            self.name: {
                "id": self.id,
                "type": self.TYPE,
                "button": {},
            },
        }

    @property
    def _prop_type(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a property type")

    @property
    def _value_for_filter(self) -> Any:  # noqa: ANN201
        raise ValueError(f"{self.__class__.__name__} doesn't need a value for filter")



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Status")


@dataclass
class Status(Property):
    status_id: str | None
    status_name: str
    status_color: str | None
    TYPE: str = "status"

    def __init__(
        self,
        name: str,
        status_name: str,
        id: str | None = None,
        status_id: str | None = None,
        status_color: str | None = None,
    ):
        self.name = name
        self.status_name = status_name
        self.id = id
        self.status_id = status_id
        self.status_color = status_color

    @classmethod
    def of(cls: Type[T], name: str, param: dict) -> T:
        return cls(
            name=name,
            status_name=param["status"]["name"],
            id=param["id"],
            status_id=param["status"]["id"],
            status_color=param["status"]["color"],
        )

    @classmethod
    def from_status_name(cls: Type[T], status_name: str, name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            status_name=status_name,
        )

    def __dict__(self):
        result = {
            "type": self.TYPE,
            "status": {
                "name": self.status_name,
            },
        }
        if self.status_id is not None:
            result["status"]["id"] = self.status_id
        if self.status_color is not None:
            result["status"]["color"] = self.status_color
        return {self.name: result}

    @property
    def _prop_type(self) -> Prop:
        return Prop.STATUS

    @property
    def _value_for_filter(self) -> Any:
        return self.status_name



from .lotion import Lotion
from .base_page import BasePage
from .decorator.notion_database import notion_database
from .decorator.notion_prop import notion_prop


__all__ = ["Lotion", "BasePage", "notion_database", "notion_prop"]



from .block import Block


class Table(Block):
    table_width: int
    has_column_header: bool
    has_row_header: bool

    def __init__(
        self,
        table_width: int,
        has_column_header: bool,
        has_row_header: bool,
        id: str,
        archived: bool,
        created_time: str,
        last_edited_time: str,
        has_children: bool,
        parent: dict,
    ):
        super().__init__(
            id, archived, created_time, last_edited_time, has_children, parent
        )
        self.table_width = table_width
        self.has_column_header = has_column_header
        self.has_row_header = has_row_header

    @staticmethod
    def of(block: dict) -> "Table":
        table = block["table"]
        table_width = table["table_width"]
        has_column_header = table["has_column_header"]
        has_row_header = table["has_row_header"]
        return Table(
            id=block["id"],
            archived=block["archived"],
            created_time=block["created_time"],
            last_edited_time=block["last_edited_time"],
            has_children=block["has_children"],
            parent=block["parent"],
            table_width=table_width,
            has_column_header=has_column_header,
            has_row_header=has_row_header,
        )

    @property
    def type(self) -> str:
        return "table"

    def to_dict_sub(self) -> dict:
        raise NotImplementedError

    def to_slack_text(self) -> str:
        return ""



from typing import Type, Any, TypeVar, cast

from ..properties.checkbox import Checkbox
from ..properties.date import Date
from ..properties.email import Email
from ..properties.multi_select import MultiSelect
from ..properties.phone_number import PhoneNumber
from ..properties.relation import Relation
from ..properties.select import Select
from ..properties.status import Status
from ..properties.text import Text
from ..properties.url import Url

from ..properties.title import Title

from ..properties.property import Property

P = TypeVar("P", bound=Property)


def __cast(value: Property, cls: Type[P]) -> P:
    parent_class = cls.__bases__[0]
    if isinstance(value, Title) and parent_class == Title:
        return cls(
            name=value.name,
            rich_text=value.rich_text,
            id=value.id,
        )
    if isinstance(value, Text) and parent_class == Text:
        return cls(
            name=value.name,
            id=value.id,
            rich_text=value.rich_text,
        )
    if isinstance(value, Checkbox) and parent_class == Checkbox:
        return cls(
            name=value.name,
            checked=value.checked,
            id=value.id,
        )
    if isinstance(value, Date) and parent_class == Date:
        return cls(
            name=value.name,
            id=value.id,
            start=value.start,
            end=value.end,
            time_zone=value.time_zone,
        )
    if isinstance(value, Email) and parent_class == Email:
        return cls(
            name=value.name,
            value=value.value,
            id=value.id,
        )
    if isinstance(value, PhoneNumber) and parent_class == PhoneNumber:
        return cls(
            name=value.name,
            value=value.value,
            id=value.id,
        )
    if isinstance(value, MultiSelect) and parent_class == MultiSelect:
        return cls(
            name=value.name,
            values=value.values,
            id=value.id,
        )
    if isinstance(value, Relation) and parent_class == Relation:
        return cls(
            name=value.name,
            id=value.id,
            id_list=value.id_list,
            has_more=value.has_more,
        )
    if isinstance(value, Select) and parent_class == Select:
        return cls(
            name=value.name,
            id=value.id,
            selected_name=value.selected_name,
            selected_id=value.selected_id,
            selected_color=value.selected_color,
        )
    if isinstance(value, Status) and parent_class == Status:
        return cls(
            name=value.name,
            id=value.id,
            status_name=value.status_name,
            status_id=value.status_id,
            status_color=value.status_color,
        )
    if isinstance(value, Url) and parent_class == Url:
        return cls(
            name=value.name,
            id=value.id,
            url=value.url,
        )
    return cast(P, value)


def notion_database(database_id: str):
    """
    クラスデコレータ: データベースIDを引数として受け取り、
    自動的に BasePage を継承させ、アノテーションの属性をプロパティ化する。
    """

    def decorator(cls):
        # 元の初期化をオーバーライドしてアノテーション属性をプロパティ化
        original_init = getattr(cls, "__init__", lambda self: None)

        def new_init(self, *args, **kwargs):
            original_init(self, *args, **kwargs)

            # クラスアノテーションに基づいてプロパティを設定
            for attr_name, attr_type in cls.__annotations__.items():

                def make_getter(name, typ: Type[P]):
                    def getter(self) -> Any:
                        # print(typ, name)  # デバッグ出力
                        result = self.get_prop(typ)  # `self.get()` は任意の実装
                        return __cast(result, typ)

                    return getter

                def make_setter(name, typ):
                    def setter(self, value: Any):
                        if not isinstance(value, typ):
                            raise TypeError(f"Expected {typ} for {name}, got {type(value)}")
                        # print(f"Setting {name} of type {typ} to {value}")  # デバッグ出力
                        self.set_prop(value)  # `set` メソッドを直接呼び出す

                    return setter

                # プロパティを作成してクラスに設定
                setattr(cls, attr_name, property(make_getter(attr_name, attr_type), make_setter(attr_name, attr_type)))

        # デコレータ引数で渡された database_id をクラス属性として設定
        setattr(cls, "DATABASE_ID", database_id)

        cls.__init__ = new_init
        cls.__module__ = cls.__module__

        return cls

    return decorator



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="MultiSelect")


@dataclass(frozen=True)
class MultiSelectElement:
    name: str
    id: str | None = None
    color: str | None = None

    def _is_value_only(self) -> bool:
        return self.id is None and self.color is None and self.name != ""

    def __dict__(self) -> dict:
        result = {
            "id": self.id,
            "name": self.name,
        }
        if self.color:
            result["color"] = self.color
        return result

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiSelectElement):
            return False
        return self.id == other.id


@dataclass
class MultiSelectElements:
    values: list[MultiSelectElement]

    def get(self, multi_select_name: str | list[str]) -> list[MultiSelectElement]:
        if isinstance(multi_select_name, str):
            return [value for value in self.values if value.name == multi_select_name]
        return [value for value in self.values if value.name in multi_select_name]

    @property
    def size(self) -> int:
        return len(self.values)


@dataclass
class MultiSelect(Property):
    values: list[MultiSelectElement]
    TYPE: str = "multi_select"

    def __init__(self, name: str, values: list[MultiSelectElement], id: str | None = None) -> None:  # noqa: A002
        self.name = name
        self.values = values
        self.id = id

    def __post_init__(self) -> None:
        if not all(isinstance(value, MultiSelectElement) for value in self.values):
            raise ValueError("All values must be MultiSelectElement instances.")

    @classmethod
    def of(cls: Type[T], name: str, param: dict) -> T:
        multi_select = [
            MultiSelectElement(
                id=element["id"],
                name=element["name"],
                color=element["color"],
            )
            for element in param["multi_select"]
        ]

        return cls(
            name=name,
            values=multi_select,
            id=param["id"],
        )

    @classmethod
    def from_name(cls: Type[T], values: list[str], name: str | None = None) -> T:
        multi_select = [MultiSelectElement(name=value) for value in values]
        return cls(
            name=name or cls.PROP_NAME,
            values=multi_select,
        )

    def _is_value_only(self) -> bool:
        for value in self.values:
            if value._is_value_only():
                return True
        return False

    def to_str_list(self) -> list[str]:
        return [value.name for value in self.values]

    @classmethod
    def create(cls: Type[T], values: list[dict[str, str]], name: str | None = None) -> T:
        """
        Create a MultiSelect instance from a list of dictionaries.

        Args:
            name (str): Name of the property.
            values (list[dict[str, str]]): List of dictionaries. Each dictionary should have keys "id" and "name".

        Returns:
            MultiSelect: MultiSelect instance.
        """
        multi_select = [MultiSelectElement(id=element["id"], name=element["name"]) for element in values]
        return cls(
            name=name or cls.PROP_NAME,
            values=multi_select,
        )

    @classmethod
    def from_elements(cls: Type[T], elements: list[MultiSelectElement], name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME, values=elements)

    def __dict__(self) -> dict:
        result = {
            "type": self.TYPE,
            "multi_select": [e.__dict__() for e in self.values],
        }
        if self.id is not None:
            result["id"] = self.id
        return {self.name: result}

    @property
    def _prop_type(self) -> Prop:
        return Prop.MULTI_SELECT

    @property
    def _value_for_filter(self) -> Any:
        if len(self.values) > 1:
            raise ValueError("MultiSelect property can only have one value for filter.")
        return self.values[0].name



import json
from unittest import TestCase

from lotion.base_page import BasePage
from lotion.properties.title import Title


class TestBasePage(TestCase):
    def test_ページを作成する(self):
        # When
        actual = BasePage.create(properties=[], blocks=[])

        # Then
        self.assertEqual([], actual.properties.values)

    def test_タイトルとリンクをSlack形式で出力する(self):
        # isinstanceのためにパスを揃える
        import sys

        sys.path.append("notion_api")

        # Given
        base_page = BasePage.create(
            properties=[Title.from_plain_text(name="名前", text="タイトル")],
        )
        base_page.update_id_and_url(page_id="dummy-id", url="http://example.com")

        # When
        actual = base_page.title_for_slack()

        # Then
        self.assertEqual("<http://example.com|タイトル>", actual)

    def test_webhookからのリクエストボディを処理できる(self):
        given = json.load(open("test/base_page_test/pattern1.json"))
        print(given)

        actual = BasePage.from_data(given)
        print(actual)
        self.assertEqual(given["id"], actual.id)

    def test_コピーを作成する(self):
        # Given
        base_page = BasePage.create(
            properties=[Title.from_plain_text(name="名前", text="タイトル")],
        )
        base_page.update_id_and_url(page_id="dummy-id", url="http://example.com")

        # When
        actual = base_page.copy()

        # Then
        self.assertIsNone(actual.id_)
        self.assertIsNone(actual.url_)
        self.assertEqual(base_page.properties.values, actual.properties.values)
        self.assertEqual(base_page.block_children, actual.block_children)
        self.assertNotEqual(base_page, actual)

    def test_オリジナルのBasePageを作成する(self):
        class OriginalPage(BasePage):
            pass

        # Given
        original_page = OriginalPage.create(
            properties=[Title.from_plain_text(name="名前", text="タイトル")],
        )
        copied_original_page = original_page.copy()

        self.assertIsInstance(original_page, OriginalPage)
        self.assertIsInstance(copied_original_page, OriginalPage)



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from ..properties.prop import Prop
from ..properties.property import Property
from .condition.cond import Cond
from .condition_ruleset import ConditionRuleset

P = TypeVar("P", bound=Property)


@dataclass(frozen=True)
class Builder:
    conditions: list[dict]

    @staticmethod
    def create() -> "Builder":
        return Builder(conditions=[])

    def add(self, prop: Property | Type[Property], cond: Cond) -> "Builder":
        if isinstance(prop, Type):
            return self._add(prop.TYPE, prop.PROP_NAME, cond)
        return self._add(prop.TYPE, prop.name, cond, prop._value_for_filter)

    def add_filter_param(self, param: dict) -> "Builder":
        return Builder(conditions=[*self.conditions, param])

    def add_created_at(self, cond_type: Cond, value: Any) -> "Builder":
        return self._add_timestamp(Prop.CREATED_TIME, cond_type, value)

    def add_last_edited_at(self, cond_type: Cond, value: Any) -> "Builder":
        return self._add_timestamp(Prop.LAST_EDITED_TIME, cond_type, value)

    def _add(self, prop_type: str, prop_name: str, cond_type: Cond, value: Any = None) -> "Builder":
        _prop_type = Prop.from_str(prop_type)
        if _prop_type == Prop.CREATED_TIME:
            raise ValueError(f"You use add_created_at() method for {prop_type}")
        if _prop_type == Prop.LAST_EDITED_TIME:
            raise ValueError(f"You use add_last_edited_at() method for {prop_type}")

        param = ConditionRuleset(_prop_type, prop_name, cond_type, value).validate().generate_param()
        return Builder(conditions=[*self.conditions, param])

    def _add_timestamp(self, prop_type: Prop, cond_type: Cond, value: Any) -> "Builder":
        param = ConditionRuleset(prop_type, "", cond_type, value).validate().generate_param()
        return Builder(conditions=[*self.conditions, param])

    def is_empty(self) -> bool:
        return len(self.conditions) == 0

    def build(self, mode: str = "and") -> dict:
        """
        :param mode: "and" or "or"

        """
        if len(self.conditions) == 0:
            raise ValueError("Filter is empty")
        if len(self.conditions) == 1:
            return self.conditions[0]
        return {
            mode: self.conditions,
        }



from dataclasses import dataclass

from .property import Property
from .title import Title


@dataclass(frozen=True)
class Properties:
    values: list[Property]

    def __post_init__(self) -> None:
        for value in self.values:
            if not isinstance(value, Property):
                msg = f"values must be Property. value: {value}"
                raise TypeError(msg)

    def __dict__(self) -> dict:
        result = {}
        for value in self.values:
            result = {**result, **value.__dict__()}
        return result

    def append_property(self, prop: Property) -> "Properties":
        props = []

        is_updated = False
        for original_prop in self.values:
            if original_prop is None:
                # この場合があるらしい。どこかで混じったかな?
                continue
            if prop.name == original_prop.name:
                props.append(prop)
                is_updated = True
            else:
                props.append(original_prop)

        # valuesにない場合は新規で追加する
        if not is_updated:
            props.append(prop)

        return Properties(values=props)

    def get_title(self) -> Title:
        for value in self.values:
            if isinstance(value, Title):
                return value
        msg = f"Title property not found. properties: {self.values}"
        raise Exception(msg)

    def get_property(self, name: str, instance_class: type) -> Property | None:
        for value in self.values:
            if isinstance(value, instance_class) and value.name == name:
                return value
        return None

    def exclude_for_update(self) -> "Properties":
        """
        更新時にエラーとなるプロパティを除外する
        """
        exclude_types = [
            "button",
            "created_by",
            "last_edited_by",
            "formula",
            "rollup",
            "unique_id",
            "files",
        ]
        print(self.values)
        print([prop.TYPE for prop in self.values])
        return Properties(values=[prop for prop in self.values if prop.TYPE not in exclude_types])

    def is_empty(self) -> bool:
        return len(self.values) == 0



from dataclasses import dataclass
from typing import Type, TypeVar

from .property import Property

T = TypeVar("T", bound="Formula")


@dataclass
class Formula(Property):
    """Formula class

    ex.
    {'id': 'h_pG', 'type': 'formula', 'formula': {'type': 'number', 'number': 50}}
    """

    _formula: dict
    TYPE: str = "formula"

    def __init__(
        self,
        name: str,
        formula: dict = {},
        id: str | None = None,  # noqa: A002
    ) -> None:
        self.name = name
        self._formula = formula
        self.id = id

    @classmethod
    def of(cls: Type[T], key: str, param: dict) -> T:
        return cls(
            id=param["id"],
            name=key,
            formula=param["formula"],
        )

    @property
    def value(self) -> dict:
        formula_type = self._formula["type"]
        return self._formula[formula_type]

    def __dict__(self) -> dict:
        raise NotImplementedError("this dict method must not be called")

    @property
    def _prop_type(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a property type")

    @property
    def _value_for_filter(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a value for filter")



from dataclasses import dataclass


@dataclass
class BaseOperator:
    id: str
    object: str

    @staticmethod
    def of(param: dict[str, str]) -> "BaseOperator":
        return BaseOperator(id=param["id"], object=param["object"])



from datetime import date, datetime, timedelta, timezone
from enum import Enum
from typing import TypeVar

JST = timezone(timedelta(hours=+9), "JST")

LENGTH_DATE = 10  # "YYYY-MM-DD"

D = TypeVar("D", bound=date)


class DateType(Enum):
    DATE = "date"
    DATETIME = "datetime"
    NONE = "none"

    @staticmethod
    def get_datetype(value: str) -> "DateType":
        try:
            # それぞれ変換できることを確認してから返却する
            if len(value) == LENGTH_DATE:
                date.fromisoformat(value)
                return DateType.DATE
            datetime.fromisoformat(value)
            return DateType.DATETIME
        except ValueError:
            return DateType.NONE


def jst_now() -> datetime:
    return datetime.now(JST)


def jst_today_datetime() -> datetime:
    return jst_now().replace(hour=0, minute=0, second=0, microsecond=0)


def jst_today(is_previous_day_until_2am: bool | None = None) -> date:
    """Return today's date in JST
    Args:
        isPreviousDayUntil2AM (bool): If True, return the previous day until 2AM. Defaults to None.
    Returns:
        date: today's date in JST
    """
    now = jst_now()
    if is_previous_day_until_2am:
        return now.date() - timedelta(days=1) if now.hour < 2 else now.date()
    return now.date()


def jst_tommorow() -> datetime:
    return jst_today_datetime() + timedelta(days=1)


def convert_to_date_or_datetime(value: str | None, cls: type[D] | None = None) -> D | None:
    if value is None:
        return None
    date_type = DateType.get_datetype(value)
    match date_type:
        case DateType.DATE:
            return _convert_date(value, cls)  # type: ignore
        case DateType.DATETIME:
            return _convert_datetime(value, cls)  # type: ignore
        case DateType.NONE:
            return None


def _convert_date(value: str, cls: type[D] | None) -> date | datetime:
    _date = date.fromisoformat(value)
    if cls is None:
        return _date
    if cls.__name__ != "datetime":
        return _date
    return datetime(_date.year, _date.month, _date.day, tzinfo=JST)


def _convert_datetime(value: str, cls: type[D] | None) -> date | datetime:
    _datetime = datetime.fromisoformat(value)
    if cls is None:
        if not __is_datatime(_datetime):
            return _datetime.date()
        return _datetime
    if cls.__name__ == "date":
        return _datetime.date()
    return _datetime


def __is_datatime(value: datetime) -> bool:
    return value.hour != 0 or value.minute != 0 or value.second != 0



def notion_prop(name: str):
    """
    クラスデコレータ: PROP_NAME を自動的に設定する。
    """

    def decorator(cls):
        # クラスに PROP_NAME を設定
        setattr(cls, "PROP_NAME", name)
        return cls

    return decorator



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Url")


@dataclass
class Url(Property):
    url: str = ""

    TYPE: str = "rich_text"
    PROP_NAME = "url"  # 自分で使うために準備。本来は不要。

    def __init__(self, name: str, url: str = "", id: str | None = None):
        self.name = name
        self.url = url
        self.id = id

    @classmethod
    def of(cls: Type[T], name: str, param: dict) -> T:
        url = param["url"] if param.get("url") else ""
        return cls(
            name=name,
            url=url,
            id=param["id"],
        )

    @classmethod
    def from_url(cls: Type[T], url: str, name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            url=url,
        )

    @classmethod
    def empty(cls: Type[T], name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME)

    def __dict__(self):
        result = {
            "type": "url",
            "url": self.url if self.url != "" else None,
        }
        if self.id is not None:
            result["id"] = self.id
        return {
            self.name: result,
        }

    @property
    def _prop_type(self) -> Prop:
        return Prop.RICH_TEXT

    @property
    def _value_for_filter(self) -> Any:
        return self.url



from dataclasses import dataclass


@dataclass
class Icon:
    type: str
    emoji: str | None = None

    @staticmethod
    def of(param: dict) -> "Icon":
        return Icon(
            type=param["type"],
            emoji=param["emoji"] if "emoji" in param else None,
        )



from dataclasses import dataclass
from typing import Type, TypeVar

from .property import Property

T = TypeVar("T", bound="UniqueId")


@dataclass
class UniqueId(Property):
    """UniqueId class

    ex.
    {'id': 'vI%3FY', 'type': 'unique_id', 'unique_id': {'prefix': None, 'number': 1}}
    """

    prefix: str | None
    number: int
    TYPE: str = "unique_id"

    def __init__(
        self,
        name: str,
        number: int,
        id: str,
        prefix: str | None = None,
    ) -> None:
        self.name = name
        self.prefix = prefix
        self.number = number
        self.id = id

    @classmethod
    def of(cls: Type[T], key: str, param: dict) -> T:
        return cls(
            id=param["id"],
            name=key,
            prefix=param["unique_id"]["prefix"],
            number=param["unique_id"]["number"],
        )

    def __dict__(self) -> dict:
        raise NotImplementedError("this dict method must not be called")

    @property
    def _prop_type(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a property type")

    @property
    def _value_for_filter(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a value for filter")



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from lotion.block.rich_text.rich_text import RichText
from lotion.block.rich_text.rich_text_builder import RichTextBuilder

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Title")


@dataclass
class Title(Property):
    # text: str
    # value: list[dict]
    rich_text: RichText
    TYPE: str = "rich_text"
    # mentioned_page_id: str | None = None
    PROP_NAME = "名前"  # 自分のデータベース用にデフォルトの名前を指定している

    def __init__(
        self,
        name: str,
        rich_text: RichText,
        id: str | None = None,
    ):
        self.name = name
        self.id = id
        self.rich_text = rich_text

    @classmethod
    def from_properties(cls: Type[T], properties: dict) -> T:
        if "Name" in properties:
            return cls.__of("Name", properties["Name"])
        if "Title" in properties:
            return cls.__of("Title", properties["Title"])
        if "名前" in properties:
            return cls.__of("名前", properties["名前"])
        msg = f"Title property not found. properties: {properties}"
        raise Exception(msg)

    @classmethod
    def from_property(cls: Type[T], key: str, property: dict) -> T:
        return cls.__of(key, property)

    def __dict__(self) -> dict:
        result: dict[str, Any] = {
            "title": self.rich_text.to_dict(),
        }
        if self.id is not None:
            result["id"] = self.id
        return {
            self.name: result,
        }

    @classmethod
    def __of(cls: Type[T], name: str, param: dict) -> T:
        rich_text = RichText.from_entity(param["title"])
        return cls(
            name=name,
            id=param["id"],
            rich_text=rich_text,
        )

    @classmethod
    def from_plain_text(cls: Type[T], text: str, name: str | None = None) -> T:
        rich_text = RichText.from_plain_text(text)
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=rich_text,
        )

    @classmethod
    def from_rich_text(cls: Type[T], rich_text: RichText, name: str | None = None) -> T:
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=rich_text,
        )

    @classmethod
    def from_mentioned_page(
        cls: Type[T],
        mentioned_page_id: str,
        prefix: str = "",
        suffix: str = "",
        name: str | None = None,
    ) -> T:
        rich_text_builder = RichTextBuilder.create()
        if prefix != "":
            rich_text_builder.add_text(prefix)
        rich_text_builder.add_page_mention(mentioned_page_id)
        if suffix != "":
            rich_text_builder.add_text(suffix)
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=rich_text_builder.build(),
        )

    @classmethod
    def from_mentioned_page_id(
        cls: Type[T],
        page_id: str,
        name: str | None = None,
    ) -> T:
        rich_text_builder = RichTextBuilder.create()
        rich_text_builder.add_page_mention(page_id)
        return cls(
            name=name or cls.PROP_NAME,
            rich_text=rich_text_builder.build(),
        )

    @property
    def text(self) -> str:
        return self.rich_text.to_plain_text()

    @property
    def _prop_type(self) -> Prop:
        return Prop.RICH_TEXT

    @property
    def _value_for_filter(self) -> Any:
        return self.text



from dataclasses import dataclass
from typing import Type, TypeVar

from .property import Property

T = TypeVar("T", bound="PhoneNumber")


@dataclass
class PhoneNumber(Property):
    """PhoneNumber class

    ex.
    {'id': 'FCsG', 'type': 'phone_number', 'phone_number': '03-1234-5678'}
    """

    value: str
    TYPE: str = "phone_number"

    def __init__(
        self,
        name: str,
        value: str = "",
        id: str | None = None,  # noqa: A002
    ) -> None:
        self.name = name
        self.value = value
        self.id = id

    @classmethod
    def of(cls: Type[T], key: str, param: dict) -> T:
        value = param.get("phone_number")
        if value is not None and not isinstance(value, str):
            raise ValueError(f"phone_number must be str, but got {type(value)}")
        return cls(id=param["id"], name=key, value=value or "")

    @classmethod
    def empty(cls: Type[T], name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME, value="")

    @classmethod
    def create(cls: Type[T], phone_number: str, name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME, value=phone_number)

    def __dict__(self) -> dict:
        result = {
            "type": self.TYPE,
            "phone_number": self.value if self.value != "" else None,
        }
        if self.id is not None:
            result["id"] = self.id
        return {
            self.name: result,
        }

    @property
    def _prop_type(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a property type")

    @property
    def _value_for_filter(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a value for filter")



from enum import Enum


class Prop(Enum):
    RICH_TEXT = "rich_text"
    CHECKBOX = "checkbox"
    DATE = "date"
    FILES = "files"
    # FORMULA = "formula"
    MULTI_SELECT = "multi_select"
    NUMBER = "number"
    PEOPLE = "people"
    RELATION = "relation"
    # ROLLUP = "rollup"
    SELECT = "select"
    STATUS = "status"
    # TIMESTAMP = "timestamp"
    ID = "unique_id"
    CREATED_TIME = "created_time"
    LAST_EDITED_TIME = "last_edited_time"
    BUTTON = "button"

    @staticmethod
    def from_str(value: str) -> "Prop":
        try:
            return Prop(value)
        except ValueError:
            raise ValueError(f"Unknown property type: {value}")



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .prop import Prop
from .property import Property

T = TypeVar("T", bound="Select")


@dataclass
class Select(Property):
    selected_name: str
    selected_id: str | None
    selected_color: str | None
    TYPE: str = "select"

    def __init__(
        self,
        name: str,
        selected_name: str = "",
        selected_id: str | None = None,
        selected_color: str | None = None,
        id: str | None = None,
    ):
        self.name = name
        self.id = id
        self.selected_name = selected_name
        self.selected_id = selected_id
        self.selected_color = selected_color

    @classmethod
    def of(cls: Type[T], name: str, param: dict) -> T:
        select = param["select"]
        if select is None:
            return cls(name=name)
        return cls(
            name=name,
            selected_id=select["id"],
            selected_name=select["name"],
            selected_color=select["color"],
            id=param["id"],
        )

    @classmethod
    def empty(cls: Type[T], name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME)

    @classmethod
    def from_name(cls: Type[T], selected_name: str, name: str | None = None) -> T:
        return cls(name=name or cls.PROP_NAME, selected_name=selected_name)

    def _is_set_name_only(self) -> bool:
        """selected_idが指定されていない場合にTrueを返す。このときは一度Selectを取得しなおす必要がある。"""
        return self.selected_id is None and self.selected_name != ""

    def is_empty(self) -> bool:
        return self.selected_name == ""

    def __dict__(self):
        if self.is_empty():
            # 空の場合はtypeだけ返す
            return {
                self.name: {
                    "type": self.TYPE,
                    "select": None,
                }
            }
        result = {
            "type": self.TYPE,
            "select": {
                "id": self.selected_id,
                "name": self.selected_name,
                "color": self.selected_color,
            },
        }
        if self.id is not None:
            result["id"] = self.id
        return {self.name: result}

    # __hash__と__eq__を実装することで、リストやセットの中で比較が可能になる
    def __hash__(self):
        return hash(self.selected_id)

    def __eq__(self, other):
        return self.selected_id == other.selected_id

    @property
    def _prop_type(self) -> Prop:
        return Prop.SELECT

    @property
    def _value_for_filter(self) -> Any:
        return self.selected_name


@dataclass(frozen=True)
class Selects:
    values: list[Select]

    def get(self, status_name: str) -> Select:
        for value in self.values:
            if value.selected_name == status_name:
                return value
        raise ValueError(f"Select not found: {status_name}")

    @property
    def size(self) -> int:
        return len(self.values)



from dataclasses import dataclass
from typing import Any, Type, TypeVar

from .property import Property

T = TypeVar("T", bound="LastEditedBy")


@dataclass
class LastEditedBy(Property):
    TYPE: str = "last_edited_by"

    def __init__(
        self,
        name: str,
        last_edited_by: dict,
        id: str | None = None,  # noqa: A002
    ) -> None:
        self.name = name
        self.last_edited_by = last_edited_by
        self.id = id

    @classmethod
    def of(cls: Type[T], key: str, param: dict) -> T:
        return cls(id=param["id"], name=key, last_edited_by=param["last_edited_by"])

    def __dict__(self) -> dict[str, Any]:
        return {
            self.name: {
                "id": self.id,
                "type": self.TYPE,
                "last_edited_by": self.last_edited_by,
            },
        }

    @property
    def _prop_type(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a property type")

    @property
    def _value_for_filter(self):
        raise ValueError(f"{self.__class__.__name__} doesn't need a value for filter")



from .block import Block


class Video(Block):
    caption: list
    external_url: str

    def __init__(
        self,
        external_url: str,
        caption: list,
        id: str | None = None,
        archived: bool | None = None,
        created_time: str | None = None,
        last_edited_time: str | None = None,
        has_children: bool | None = None,
        parent: dict | None = None,
    ):
        super().__init__(
            id, archived, created_time, last_edited_time, has_children, parent
        )
        self.external_url = external_url
        self.caption = caption

    @staticmethod
    def of(block: dict) -> "Video":
        video = block["video"]
        video_external = video["external"] if "external" in video else {}
        return Video(
            id=block["id"],
            archived=block["archived"],
            created_time=block["created_time"],
            last_edited_time=block["last_edited_time"],
            has_children=block["has_children"],
            parent=block["parent"],
            caption=video["caption"],
            external_url=video_external["url"] if "url" in video_external else "",
        )

    @staticmethod
    def from_external_url(url: str) -> "Video":
        return Video(
            external_url=url,
            caption=[],
        )

    @property
    def type(self) -> str:
        return "video"

    def to_dict_sub(self) -> dict:
        return {
            "caption": self.caption,
            "type": "external",
            "external": {"url": self.external_url},
        }

    def to_slack_text(self) -> str:
        return self.external_url
